/* 
 * File:   main.c
 * Authors: Michael Kay, Chris Mulholland, Eric Fernandez Salas,
 *          Michael Kercher, and Fernanda Urteaga Zambrano
 *
 * Created on April 28, 2019, 9:20 AM
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <xc.h>
#include <sys/attribs.h>
#include <sys/kmem.h>
#include "FT8xx.h"
#include "Library.h"
#include "EVE_RGB565.h"
#include "DIGITfont.h"
#include "LeaveItToEVE.h"
#include "EVE_PNG.h"

/* 
 * 
 * 
 * 
 * Function Declarations 
 * 
 * 
 * 
 * 
 * 
 */

// General Initialization //
void initInterrupts(void);
void initIO(void);
unsigned short convert_to_analog(unsigned short i);
void initadc(void);

// Interrupt Timers //
void initTimer2(void);
void initTimer3();

// SPI //
void write_to_GDAC(int, int);
void write_to_LDAC(int, int);
void write_to_RDAC(int, int);
void initDAC_SPI2(void);
void initENCODER_SPI2(void);
void initISO_SPI2(void);

short WriteReadSPI(unsigned short i);
short ReadSPI(unsigned short i);

// CAN //
void dotheCAN();
void initCAN2();
void set_Baud();
void initFIFO();
void Read_FIFO();
void ProcessMessage1();
void ProcessMessage2();
void initFilter();

// Message Processing //
void whichKelly(void);
void computeDynamics(void);
void decodeErrors(void);

// UART //
void configUART2();
void tputc(unsigned char);
void printData();
void printRF();

// LCD //
void initLCD();
void APP_Init(void);
void APP_Gauge(void);
void updateScreen();
void errorMessageTranslator(unsigned short);
void kill1Time();
void suchTime();
void getInfo();

/* 
 * 
 * 
 * 
 * 
 * Define Statements
 * 
 * 
 * 
 * 
 * 
 */

// Chip Select Pins
#define CS_ROT LATEbits.LATE7   // Rotary Encoder CS
#define CS_G LATEbits.LATE6     // GEN DAC CS
#define CS_L LATGbits.LATG9     // Left DAC CS
#define CS_R LATFbits.LATF3     // Right DAC CS
#define CS_LCD LATDbits.LATD4   // LCD CS
#define CS_ISOB LATEbits.LATE5  // ISO SPI CS

//timer 4 interrupt flag for PID
#define T4IntFlag IFS0bits.T4IF

//define pins to read inputs from for this drive system madness
#define DIRECTION_IN 2                  // define analog pin RB2/AN2 to track car direction input (fwd/rev/neutral)
#define DIRECTION_OUT PORTEbits.RE3     // define pin RE3 to track direction output 
#define AIR_INTERRUPT PORTCbits.RC13    // define pin RC13 as reading AIR relay state
#define THROTTLE1 4                     // define pin RB4/AN4 as reading throttle1 signal
#define THROTTLE2 5                     // define pin RB5/AN5 as reading throttle2 signal
#define BRAKE 3                         // define pin RB3/AN3 as reading brake signal

//experimentally determined voltage minimums and maximums for throttle, brake, and Kelly 
//ONLY CHANGE THESE DEFINE STATEMENTS EVERYTHING ELSE IS DERIVED FROM THEM//
#define THROTTLE1_V_OUT_MIN 1.6    // experimentally determined minimum output allowed for throttle port 1 in V
#define THROTTLE1_V_OUT_MAX 4.1   // experimentally determined maximum output allowed for throttle port 1 in V
#define THROTTLE2_V_OUT_MIN 0.8    // experimentally determined minimum output allowed for throttle port 2 in V
#define THROTTLE2_V_OUT_MAX 4.1    // experimentally determined maximum output allowed for throttle port 2 in V
// set brake values equal to throttle1 values 
#define BRAKE_V_OUT_MIN 1.6         // experimentally determined minimum output allowed for brake in V (same as T1)
#define BRAKE_V_OUT_MAX 4.1        // experimentally determined maximum output allowed for brake in V (same as T1)
//voltage values corresponding to the min and max Kelly signals allowed (on a 5V scale) - only change these
#define KELLY_DEAD_V_MIN 0.75       // voltage value of minimum Kelly signal allowed (on 5V scale)
#define KELLY_DEAD_V_MAX 4.25       // voltage value of maximum Kelly signal allowed (on 5V scale)

// max difference between readings of throttle ports (5%), mapped to 1024 scale
#define THROTTLE_DIFF 700           // = 0.171*4096

//Kelly min and max voltage values mapped onto a 0-4096 scale
#define KELLY_DEAD_MIN 614  // = (KELLY_DEAD_V_MIN/5)*4096 , value of max Kelly signal allowed (4096 scale) - final value to use
#define KELLY_DEAD_MAX 3482 // = (KELLY_DEAD_V_MAX/5)*4096 , value of max Kelly signal allowed (4096 scale) - final value to use
#define KELLY_DIFF (KELLY_DEAD_MAX-KELLY_DEAD_MIN) //use this to make scale start at 0 (with this as max value)

//map throttle1 voltage to 1023 scale, determine constant to map signal to Kelly input range
#define THROTTLE1_OUT_MIN 279 //298 // = (THROTTLE1_V_OUT_MIN*(3/5)*(1023/3.3)) , throttle1 minimum allowed value on 1023 scale
#define THROTTLE1_OUT_MAX 785 // = (THROTTLE1_V_OUT_MAX*(3/5)*(1023/3.3))throttle1 maximum allowed value on 1023 scale
#define THROTTLE1_DIFF (THROTTLE1_OUT_MAX-THROTTLE1_OUT_MIN)     //difference of min and max throttle1 values (adjust mapping to start at 0)
//#define THROTTLE1_CONST (KELLY_DIFF/THROTTLE1_DIFF)              //constant value to convert to scale needed for Kelly signal input
#define THROTTLE1_CONST 5.668 //5.889              //constant value to convert to scale needed for Kelly signal input

//map throttle2 voltage to 1023 scale, determine constant to map signal to Kelly input range
//#define THROTTLE2_OUT_MIN (THROTTLE2_V_OUT_MIN*(3/5)*(1023/3.3)) //throttle2 minimum allowed value on 1023 scale
//#define THROTTLE2_OUT_MAX (THROTTLE2_V_OUT_MAX*(3/5)*(1023/3.3)) //throttle2 maximum allowed value on 1023 scale
#define THROTTLE2_OUT_MIN 149 //throttle2 minimum allowed value on 1023 scale
#define THROTTLE2_OUT_MAX 785 //throttle2 maximum allowed value on 1023 scale
#define THROTTLE2_DIFF (THROTTLE2_OUT_MAX-THROTTLE2_OUT_MIN)     //difference of min and max throttle2 values (adjust mapping to start at 0)
//#define THROTTLE2_CONST (KELLY_DIFF/THROTTLE2_DIFF)              //constant value to convert to scale needed for Kelly signal input
#define THROTTLE2_CONST 4.509            //constant value to convert to scale needed for Kelly signal input

//map brake voltage to 1023 scale, determine constant to map signal to Kelly input range
#define BRAKE_OUT_MIN 279 //298   // = (BRAKE_V_OUT_MIN*(3/5)*(1023/3.3)) , brake minimum allowed value on 1023 scale
#define BRAKE_OUT_MAX 785   // = (BRAKE_V_OUT_MAX*(3/5)*(1023/3.3)) , brake maximum allowed value on 1023 scale
#define BRAKE_DIFF (BRAKE_OUT_MAX-BRAKE_OUT_MIN)         //difference of min and max throttle2 values (adjust mapping to start at 0)
#define BRAKE_CONST 5.668 //5.889   // = (KELLY_DIFF/BRAKE_DIFF) , constant value to convert to scale needed for Kelly signal input
#define REGEN_LEVEL 2100 //amount of regen

//voltage values for forward and reverse states ranges surrounding ((2/3*Vdd) and (1/3*Vdd))
#define FWD_V_MIN 1.98       //minimum voltage value for range of signal designated to FWD state 0.6*Vdd                
#define FWD_V_MAX 2.64       //maximum voltage value for range of signal designated to FWD state 0.8*Vdd
#define REV_V_MIN 0.66       //minimum voltage value for range of signal designated to REV state 0.2*Vdd
#define REV_V_MAX 1.32       //maximum voltage value for range of signal designated to REV state 0.4*Vdd
//mapped values for forward and reverse states from 3.3V scale to 0-1023 scale
#define FWD_MIN 613.8  // = (FWD_V_MIN/3.3)*1023 , minimum 1023 value for range of signal designated to FWD state 0.6*Vdd            
#define FWD_MAX 818.4  // = (FWD_V_MAX/3.3)*1023 , maximum 1023 value for range of signal designated to FWD state 0.8*Vdd
#define REV_MIN 204.6  // = (REV_V_MIN/3.3)*1023 , minimum 1023 value for range of signal designated to REV state 0.2*Vdd
#define REV_MAX 409.2  // = (REV_V_MAX/3.3)*1023 , maximum 1023 value for range of signal designated to REV state 0.4*Vdd

//torque vectoring constant
//NOT YET DEFINED//
#define TORQUE_VECTORING_CONSTANT 1
// define temperature for overheating to trigger warning for motor or controller, read in from thermistors
//NOT YET DEFINED//
#define MAX_TEMP 0

// STATE DIAGRAM DEFINE STATEMENTS //

//define drive states 
#define Wait 0
#define Precharge 1
#define Neutral 2
#define Fwd 3
#define Left 4
#define Right 5
#define Rev 6
//define charge states
#define Charge 0
#define Idle 1
//define on/off states
#define on 1
#define off 0

//define value for minimum voltage charge
#define MinV 60
//define value for maximum voltage charge
#define MaxV 150

//define pins
#define CHARGE_ENABLE PORTDbits.RD6
#define ENABLE PORTDbits.RD7
#define PRECHARGE_CONTACTOR LATEbits.LATE2
#define DISCHARGE_CONTACTOR LATDbits.LATD11
#define IGNITION LATBbits.LATB6

// Defines for ADC Conversion Functions //
#define SAMP AD1CON1bits.SAMP // bit that clears when sampling is terminated
#define ADCDONE AD1CON1bits.DONE

// Defines for LCD Functions //
#define LCD_800_480				
#define FT81x_ACTIVE	0x00


/* 
 * 
 * 
 * 
 * Global Variables 
 * 
 * 
 * 
 * 
 * 
 */

//########################## Rotary Encoder Variables #######################//

// define variables for SPI communication
    int readbyte = 0x00;
    int readMSB = 0x00;
    int readLSB = 0x00;
    int Encodertransaction=1;
// define variables to make conversion to angle
    float steering_angle = 0;
    int my_MSB = 0x0000;
    int my_LSB = 0x0000;
    int my_Position = 0x0000;
    int degrees = 360;
    int positions = 4096;
// define variable to hold left/right turn state
    int turn_state = 0;
// define variable for switch statement/interrupt
    int read_state = 0;
// define variable to track first time setting position
    int first_time = 1;
    
// indicator for SPI switching
    int SPI_init_done=1;
    
// Global for direction state printing to LCD
    
    
//########################### CAN Variables ###########################//    

/* Data structure for CAN full receive message buffer.*/
    
/* Define the sub-components of the data structure as specified in Table 34-5 */
/* Create a CMSGSID data type. */
    typedef struct
    {
        unsigned SID:11;
        unsigned FILHIT:5;
        unsigned CMSGTS:16;
    }rxcmsgsid;
/* Create a CMSGEID data type. */
    typedef struct
    {
        unsigned DLC:4;
        unsigned RB0:1;
        unsigned :3;
        unsigned RB1:1;
        unsigned RTR:1;
        unsigned EID:18;
        unsigned IDE:1;
        unsigned SRR:1;
        unsigned :2;
    }rxcmsgeid;
/* Create a CMSGDATA0 data type. */
    typedef struct
    {
        unsigned Byte0:8;
        unsigned Byte1:8;
        unsigned Byte2:8;
        unsigned Byte3:8;
    }rxcmsgdata0;
/* Create a CMSGDATA1 data type. */
    typedef struct
    {
        unsigned Byte4:8;
        unsigned Byte5:8;
        unsigned Byte6:8;
        unsigned Byte7:8;
    }rxcmsgdata1;
    
/* This is the main data structure. */
    typedef union uCANRxMessageBuffer {
    struct
        {
            rxcmsgsid CMSGSID;
            rxcmsgeid CMSGEID;
            rxcmsgdata0 CMSGDATA0;
            rxcmsgdata0 CMSGDATA1;
    };
        int messageWord[4];
    }CANRxMessageBuffer;
    
    volatile int EID = 0;
    volatile float rpm = 0;
    volatile float rpm_last = 0;
    volatile float rpm_new = 0;
    volatile float avg_rpm = 0;
    volatile float speed = 0;
    volatile float power = 0;
    volatile float motor_current = 0;
    volatile float mot_G = 0;
    volatile float mot_last = 0;
    volatile float mot_new = 0;
    volatile int bat_voltage = 0;
    volatile int error_lsb = 0;
    volatile int error_msb = 0;
    volatile int contr_temp = 0;
    volatile int contr_temp_G = 0;
    volatile int contr_temp_LR = 0;
    volatile int motor_temp = 0;
    volatile int mot_temp_G = 0;
    volatile int mot_temp_LR = 0;
    volatile int contr_status = 0;
    volatile int switch_status = 0;
    volatile float throttle = 0;
    
    volatile char* msg_array[8];
    
//######################### Drive States Variables #########################//

//initialize state variable(s)
    int drive_state = 0;  //current state in switch statement
    char* drive_str = "";
    char* charge_str = "";    
    int charge_state = 0;

//KERCHER CODE FOR RPM AND REGEN
    int RPM_PID = 0;
    int Regen_PID = 0;
    int Servo = 0;
    
//ERIC'S CODE FOR ERRORS, MUST SEND MESSAGE TO SLAVE PIC EACH TIME TO AS WELL AS HERE
    int Error = 0;

//INTEGRAT ERICS AMS AND LCD CODE for these values
    int SumV = 0;

//Keep track of time caps are charging
    int T_charge = 0;

// Analog input variables
    unsigned short AN_BRAKE = 0;
    unsigned short AN_THROTTLE1 = 0;
    unsigned short AN_THROTTLE2 = 0;
    unsigned short AN_DIRECTION_IN = 0;
    
// Current Variables for Throttle and Brake (range 0-930 b/c voltage divider input signal to board is 0-3V)
    int throttle1_val = 0;
    int throttle2_val = 0;
    int throttle_val_final = 0;
    int brake_val_final = 0;
//Torque Vectoring Variables
    int Throttle_L = 0;
    int Throttle_R = 0;
    int Brake_L = 0;
    int Brake_R = 0;
    int Rev_L = 0;
    int Rev_R = 0;
//state variable for possible direction states (0 is neutral)
    int direction_state = 0;
    
// CAN messager Error handling //
    unsigned char CAN_PROCESSED = 1;
    
//######################### Charge State Variables #########################//
    
volatile int CN_count=0; // global for CN toggle
volatile int oldRPM =0, desiredRPM=6000, newRPM=0, error;
unsigned int integral=0;
float Kp = 0.12, Ki=.16, Kd=.01; //PID settings 
int PID, derivative;
int servoMin=5300; //Servo position at idle
int servoMax=3320; //Servo position at full throttle
    
//######################### AMS & LCD Variables #########################//
    
unsigned short temp=0, val=0;
float cellVolt=0;
unsigned short sumAll=0;
unsigned short sensorNum = 0xFF;
unsigned short overvoltNum = 0xFF;
unsigned short sensorVal=0;
char* AMSboard = "No Error";
char* direction_state_LCD = "Wait";
char CAN_ER[16] = " ";
char *errorPointer[3];
char CAN_COUNT = 0;
int CAN_error_buffer=0;
unsigned char new_CAN_Error_MSG = 0;
unsigned short tempFinal;
unsigned int percBat = 0;

// LCD display parameters
uint16_t lcdWidth;                                                              // Active width of LCD display
uint16_t lcdHeight;                                                             // Active height of LCD display
uint16_t lcdHcycle;                                                             // Total number of clocks per line
uint16_t lcdHoffset;                                                            // Start of active line
uint16_t lcdHsync0;                                                             // Start of horizontal sync pulse
uint16_t lcdHsync1;                                                             // End of horizontal sync pulse
uint16_t lcdVcycle;                                                             // Total number of lines per screen
uint16_t lcdVoffset;                                                            // Start of active screen
uint16_t lcdVsync0;                                                             // Start of vertical sync pulse
uint16_t lcdVsync1;                                                             // End of vertical sync pulse
uint8_t lcdPclk;                                                                // Pixel Clock
uint8_t lcdSwizzle;                                                             // Define RGB output pins
uint8_t lcdPclkpol;                                                             // Define active edge of PCLK

uint32_t ramDisplayList = RAM_DL;                                               // Set beginning of display list memory
uint32_t ramCommandBuffer = RAM_CMD;                                            // Set beginning of graphics command memory

uint16_t FT81x_GPIO_DAT;                                                        // Used for GPIO registers
uint16_t FT81x_GPIO_DIR;
uint8_t dly = 0;
uint8_t PWM = 0;
uint16_t endSum=0;

uint32_t dlOffset = 0;

/*
 * 
 * 
 * 
 * Begin LCD Initialization Code
 * 
 * 
 * 
 * 
 */

void APP_Init(void)
{
    // ----------------------- Cycle PD pin to reset device --------------------
    
    MCU_PDlow();                                                                // PD low to reset device                                                                
    MCU_Delay_20ms();
    MCU_PDhigh();                                                               // PD high again
    MCU_Delay_20ms();

    // ---------------------- Delay to allow FT81x start-up --------------------    
    
    EVE_CmdWrite(FT81x_ACTIVE, 0x00);                                           // Sends 00 00 00 to wake FT8xx
    
    MCU_Delay_500ms();                                                          // 500ms delay (EVE requires at least 300ms here))
        
    // --------------- Check that FT8xx ready and SPI comms OK ----------------- 
    
    while (EVE_MemRead8(REG_ID) != 0x7C);                                        // Read REG_ID register until reads 0x7C
   
      
    while (EVE_MemRead8(REG_CPURESET) != 0x00);                                  // Ensure CPUreset register reads 0 and so FT8xx is ready   

    // ---------------- Configure the GPIO and PWM  --------------------
    
    EVE_MemWrite8(REG_PWM_DUTY, 0);                                             // Backlight off
   
    FT81x_GPIO_DAT = EVE_MemRead16(REG_GPIOX);                                  // Read the FT81x GPIO register for a read/modify/write operation
    FT81x_GPIO_DAT = FT81x_GPIO_DAT | 0x8000;                                   // set bit 15 of GPIO register (DISP). If using audio, also enable the amplifier shutdown GPIO line here
    EVE_MemWrite16(REG_GPIOX, FT81x_GPIO_DAT);                                  // Enable the DISP signal to the LCD panel

    FT81x_GPIO_DIR = EVE_MemRead16(REG_GPIOX_DIR);                              // Read the FT81x GPIO DIR register for a read/modify/write operation
    FT81x_GPIO_DIR = FT81x_GPIO_DIR | 0x8000;                                   // DISP is output, if using audio, also set the associated shutdown line as output (check your Eval module schematic))
    EVE_MemWrite16(REG_GPIOX_DIR, FT81x_GPIO_DIR);                              // Enable the DISP signal to the LCD panel
        
    // ------------------------- Display settings ------------------------------

    // WQVGA display parameters
    lcdWidth   = 800;                                                           // Active width of LCD display
    lcdHeight  = 480;                                                           // Active height of LCD display
    lcdHcycle  = 928;                                                           // Total number of clocks per line
    lcdHoffset = 88;                                                            // Start of active line
    lcdHsync0  = 0;                                                             // Start of horizontal sync pulse
    lcdHsync1  = 48;                                                            // End of horizontal sync pulse
    lcdVcycle  = 525;                                                           // Total number of lines per screen
    lcdVoffset = 32;                                                            // Start of active screen
    lcdVsync0  = 0;                                                             // Start of vertical sync pulse
    lcdVsync1  = 3;                                                             // End of vertical sync pulse
    lcdPclk    = 2;                                                             // Pixel Clock
    lcdSwizzle = 0;                                                             // Define RGB output pins
    lcdPclkpol = 1;                                                             // Define active edge of PCLK
       
    EVE_MemWrite16(REG_HSIZE,   lcdWidth);	
    EVE_MemWrite16(REG_HCYCLE,  lcdHcycle);	
    EVE_MemWrite16(REG_HOFFSET, lcdHoffset);	
    EVE_MemWrite16(REG_HSYNC0,  lcdHsync0);	
    EVE_MemWrite16(REG_HSYNC1,  lcdHsync1);	
    EVE_MemWrite16(REG_VSIZE,   lcdHeight);	
    EVE_MemWrite16(REG_VCYCLE,  lcdVcycle);	
    EVE_MemWrite16(REG_VOFFSET, lcdVoffset);	
    EVE_MemWrite16(REG_VSYNC0,  lcdVsync0);	
    EVE_MemWrite16(REG_VSYNC1,  lcdVsync1);	
    EVE_MemWrite8(REG_SWIZZLE,  lcdSwizzle);	
    EVE_MemWrite8(REG_PCLK_POL, lcdPclkpol);
    
    // ---------------------- Create an initial screen before we enable the display -------------------------

    ramDisplayList = RAM_DL;                                                    // start of Display List
    EVE_MemWrite32(ramDisplayList, 0x02000000);                                 // Clear Color RGB sets the colour to clear screen to

    ramDisplayList += 4;                                                        // point to next location
    EVE_MemWrite32(ramDisplayList, (0x26000000 | 0x00000007));                  // Clear 00100110 -------- -------- -----CST  (C/S/T define which parameters to clear)

    ramDisplayList += 4;                                                        // point to next location
    EVE_MemWrite32(ramDisplayList, 0x00000000);                                 // DISPLAY command 00000000 00000000 00000000 00000000 (end of display list)

    EVE_MemWrite32(REG_DLSWAP, DLSWAP_FRAME);                                   // Swap display list to make the edited one active

    // -------------------- Now turn on PCLK and ramp the PWM up  -------------------------------------
    // ---------------- This provides a clean start-up for the application ----------------------------
        
    EVE_MemWrite8(REG_PCLK, lcdPclk);                                           // Now start clocking data to the LCD panel

    MCU_Delay_20ms();
    
    for(PWM = 0; PWM <= 127; PWM ++)
    {
        EVE_MemWrite8(REG_PWM_DUTY, PWM);      
        MCU_Delay_20ms();
    }
   
}

void APP_Gauge(void)
{       
    
    uint8_t GaugeVal = 0;
    uint8_t GaugeDir = 0;
    uint16_t Toggle = 0;
    uint16_t GaugeVal2 = 0;
    uint16_t GaugeDir2 = 0;
    uint16_t BatVal = 0;
    uint8_t Bar2 = 0;
    
    //Right side
    uint16_t TempVal = 100;
    uint16_t OVflagVal = 0;
    uint16_t ThrottleVal = 0;
    uint16_t AngleVal = 0;
    uint16_t VoltVal = 0;
    uint16_t ICRPM = 0;
    uint8_t percICRPM = 0;
    
    // *******************  Create the static screen content *******************
        
    API_LIB_BeginCoProList();                                               // Begin new screen
    API_CMD_DLSTART();                                                      // Tell co-processor to create new Display List
    API_CMD_GRADIENT(300, 0, 0x0000, 300, 480, 0x0000);                     // Gradient background
    
    API_CLEAR_COLOR_RGB(0,0,0);
    API_COLOR_RGB(200, 200, 200);        
    API_BEGIN(LINE_STRIP);                                                      // Draw line strip in grey color to make a border around the screen
    API_VERTEX2F(20*16, 20*16);                                                 // Top-left coordinate
    API_VERTEX2F(780*16, 20*16);                                                // Top-Right corner
    API_VERTEX2F(780*16, 460*16);                                               // Bottom-Right corner
    API_VERTEX2F(20*16, 460*16);                                                // Bottom-left corner 
    API_VERTEX2F(20*16, 20*16);                                                 // Back up to the top-left corner to complete the border
    API_END();                                                                  // Finish drawing line strip
    API_LIB_EndCoProList();                                                     // Finish the co-processor list burst write
    API_LIB_AwaitCoProEmpty();                                                  // Wait until co-processor has consumed all commands and generated DL
    
    // ******************  Store the static screen content *********************
    // Determine length to be copied and then copy the generated DL to RAM_G 
    // by running a MEMCPY command
    
    dlOffset = EVE_MemRead32(REG_CMD_DL);                                       // Check where the end of the new DL is in RAM_DL (i.e. the length))
    
    API_LIB_BeginCoProList();                                                   // Begin new co-processor command list
    API_CMD_MEMCPY(1000, RAM_DL, dlOffset);                                     // Copy [dlOffset] bytes beginning from RAM_DL, to RAM_G location 1000 onwards  
    API_LIB_EndCoProList();                                                     // Finish the co-processor list burst write
    API_LIB_AwaitCoProEmpty();                                                  // Wait until co-processor has consumed all commands
        
    // *********  Create the final screen adding the dynamic parts *************
    updateScreen();
    
}

int keep_AMS_error=0;
char* keep_AMS_message;
int keep_sensor_value;
char* charge_state_LCD = "No state";
void updateScreen(){
    
// Left Side Variables //
    int GaugeVal = power-.5;            // Right gauge
    uint8_t GaugeDir = 0;
    uint16_t Toggle = 0;
    uint16_t GaugeVal2 = speed-.5;         // Left gauge
    uint16_t GaugeDir2 = 0;
    uint16_t BatVal = 0;
    uint16_t ICRPM = 0;       // This will be the IC engine RPM
    uint8_t percICRPM = ((float) newRPM / 13000) * 100;
    
// Right Side Variables //
    uint16_t TempVal = 100;
    uint16_t OVflagVal = 0;
    uint16_t ThrottleVal = throttle-.5;
    uint16_t AngleVal = 0;
    uint16_t VoltVal = 0;
    API_LIB_BeginCoProList();           // Begin new screen
    API_CMD_DLSTART();                  // Tell co-processor to create new Display List
    API_CLEAR_COLOR_RGB(255, 255, 255); // Specify color to clear screen to
    API_CLEAR(1,1,1);                   // Clear color, stencil, and tag buffer
    API_CMD_APPEND(1000, dlOffset);
    API_COLOR_RGB(255, 255, 255);       // Set color to blue and write some text

// Left Side Print //
    uint16_t batteryVal = endSum;
    percBat = (batteryVal-44)*100/(150-44);
    API_CMD_GAUGE(125, 180, 100, 0, 4, 8, GaugeVal2, 80);                    // Draw gauge with pointer defined by variable at top of this function 
    API_CMD_GAUGE(335, 180, 100, 0, 4, 8, GaugeVal+250, 500);
    API_CMD_FGCOLOR(0xFF0000);   
    API_CMD_NUMBER(123, 245, 25, OPT_CENTER, GaugeVal2);                     // Print the numerical value of the variable GaugeVal
    
    GaugeVal = abs(GaugeVal);
    API_CMD_NUMBER(333, 245, 25, OPT_CENTER, GaugeVal);
    
    API_CMD_TEXT(295, 45, 24, 0, "Power");
    API_CMD_TEXT(95, 45, 24, 0, "MPH"); // First goes left to right // Second goes bottom to top
    // First Bar //
    API_CMD_PROGRESS(70, 340, 330, 23, 0, percBat, 100);
    API_CMD_TEXT(115, 310, 24, 0, "Battery          %");
    API_CMD_NUMBER(230, 325, 24, OPT_CENTER, percBat);
    // Second Bar //
    API_CMD_PROGRESS(70, 410, 330, 23, 0, newRPM, 13000);
    API_CMD_TEXT(115, 380, 24, 0, "RPM      ");
    API_CMD_NUMBER(230, 395, 24, OPT_CENTER, newRPM);

// Right Side Print //
    API_CMD_TEXT(485, 115, 24, 0, "Temp. = ");
    int tempAvg = ((int) tempFinal) - 293;
    API_CMD_NUMBER(580, 115, 24, 256, tempAvg);
    API_CMD_TEXT(485, 75, 24, 0, "Error = ");
    API_CMD_NUMBER(570, 85, 24, OPT_CENTER, Error);
    API_CMD_TEXT(550, 30, 30, 0, direction_state_LCD);//direction_state_LCD
    
    API_CMD_TEXT(485, 160, 24, 0, "S.W Angle = ");
    API_CMD_NUMBER(640, 175, 24, OPT_CENTER, steering_angle);
    API_CMD_TEXT(485, 190, 24, 0, "S. Voltage = ");
    API_CMD_NUMBER(640, 205, 24, OPT_CENTER, batteryVal);
    API_CMD_TEXT(485, 220, 24, 0, "Charge St. = ");
    API_CMD_TEXT(640, 220, 24, 0, charge_state_LCD);
    API_CMD_TEXT(485, 280, 24, 0, "Errors : ");
    int pos=340;
    
    if(((AMSboard == "No Error") && new_CAN_Error_MSG == 0) && !keep_AMS_error){   // No CAN Error and No AMS Error
        API_CMD_TEXT(485, 310, 24, 0, AMSboard);
    }else if(((AMSboard != "No Error") && (new_CAN_Error_MSG == 0)) || keep_AMS_error){  // AMS Error and No CAN Error
        if(!keep_AMS_error){
        keep_AMS_message = AMSboard;
        keep_sensor_value = sensorVal;
        } 
        API_CMD_TEXT(485, 310, 24, 0, keep_AMS_message);
        API_CMD_TEXT(485, pos, 24, 0, "Sensor: ");
        API_CMD_NUMBER(600,pos, 24,0,keep_sensor_value);
        pos=pos+30;
        keep_AMS_error=1;
        Error=1;
    }else if((AMSboard == "No Error") && new_CAN_Error_MSG == 1){   // CAN Error and No AMS Error
        CAN_error_buffer = CAN_COUNT;
        while(CAN_error_buffer > 0){
        API_CMD_TEXT(485, pos+60-CAN_error_buffer*30, 24, 0, errorPointer[CAN_error_buffer-1]);
        CAN_error_buffer--;
        }
    }else if((AMSboard != "No Error") && new_CAN_Error_MSG == 1){   // CAN Error and AMS Error
        API_CMD_TEXT(485, 310, 24, 0, AMSboard);
        API_CMD_TEXT(485, pos, 24, 0, "Sensor: ");
        API_CMD_NUMBER(600,pos, 24,0,sensorVal);
        pos=pos+30;
        CAN_error_buffer = CAN_COUNT;
        while(CAN_error_buffer > 0){
        API_CMD_TEXT(485, pos+60-CAN_error_buffer*30, 24, 0, errorPointer[CAN_error_buffer-1]);
        CAN_error_buffer--;
        keep_AMS_error=1;
        }
        Error=1;
    }
    
    API_DISPLAY();                                                          // Tell EVE that this is end of list
    API_CMD_SWAP();                                                         // Swap buffers in EVE to make this list active
    API_LIB_EndCoProList();                                                 // Finish the co-processor list burst write
    API_LIB_AwaitCoProEmpty();                                              // Wait until co-processor has consumed all commands
    
}

void initLCD(){
    MCU_Init();         // MCU Initialization including I/O, oscillator and SPI
    APP_Init();         // EVE initialization
    APP_Gauge();
    updateScreen();
}

/*
 * 
 * 
 * 
 * AMS Communications Code
 * 
 * 
 * 
 * 
 */

void errorMessageTranslator(unsigned short x){
    unsigned short type = 0b11000000;
    sensorNum = (0xFF & x);
    overvoltNum = (0xFF & x);
    AMSboard = " ";
  
    type = (type & x);
// Error type is Temperature is out of range
    if(type == 0x80){
        sensorNum = (sensorNum & 0b00111111);
        if(sensorNum < 8){
            AMSboard = "Board A - Temperature Error";
            
        }else if(sensorNum < 16){
            AMSboard = "Board B - Temperature Error";
            sensorNum = sensorNum - 8;
        }else if(sensorNum < 24){
            AMSboard = "Board C - Temperature Error";
            sensorNum = sensorNum - 16;
        }else{
            AMSboard = "Board D - Temperature Error";
            sensorNum = sensorNum - 24;
        }
        sensorVal = sensorNum;
    }else if(type == 0xC0){  // Error type is Over voltage is triggered
        overvoltNum = (overvoltNum & 0b00111111);
        if(overvoltNum < 15){
            AMSboard = "Board A - OV Error";
        }else if(overvoltNum < 30){
            AMSboard = "Board B - OV Error";
            overvoltNum = overvoltNum - 15;
        }else if(overvoltNum < 45){
            AMSboard = "Board C - OV Error";
            overvoltNum = overvoltNum - 30;
        }else{
            AMSboard = "Board D - OV Error";
            overvoltNum = overvoltNum - 45;
        }
        sensorVal = overvoltNum;
    }else{
        AMSboard = "No Error";
        sensorVal=0;
    }
    
}

void getInfo(){
    suchTime();
    unsigned short temp;
    CS_ISOB=0;
    temp = WriteReadSPI(0xFA); // Start :: Next variable is high byte of Sum of cells
    CS_ISOB=1;
    kill1Time();
    CS_ISOB=0;
    sumAll = (WriteReadSPI(0xFB) << 8); // Ask for next variable :: low byte of Summ of cells
    CS_ISOB=1;
    kill1Time();
    CS_ISOB=0;
    sumAll = (sumAll & 0xFF00) + WriteReadSPI(0xFC);
    CS_ISOB=1;
    kill1Time();
    CS_ISOB=0;
    temp = WriteReadSPI(0xFD); // This is the error message
    CS_ISOB=1;
    kill1Time();
    errorMessageTranslator(temp);
    CS_ISOB=0;
    tempFinal = WriteReadSPI(0xFE) + 200; // This is the error message
    CS_ISOB=1;
    
    // Implement after thermistor are installed //
    
//    CS_ISOB=0;
//    temp = WriteReadSPI(0xFE); // This is the error message
//    CS_ISOB=1;

}
 
void kill1Time(){
    int x=0;
    while(x<100){
        x++;
    }
}

void suchTime(){
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    kill1Time();
    
}

/* 
*
*
*
*
* Charging State/Engine Control Code
*
*
*
*
*/

void init_OC(void){
    // Servo is a 1 to 2 ms pulse and a 20 ms period. OCR register if 50K is 20 ms, then 5000 is 2 ms, and 2500 is 1. Set OCTSEL to correct timer base
    // INTEnableSystemMultiVectoredInt(); // Enable system wide interrupt to multivectored mode.
    OC1CON = 0x0000;            // Turn off the OC1 when performing the setup
    PR2 = 49999;                // Set period
    OC1RS = 5300;  // Initialize secondary Compare register
    OC1R = 5300;          // Initialize primary Compare register
    OC1CONbits.OCM = 0b110;            // Configure for PWM mode without Fault pin enabled
    OC1CONbits.OC32 = 0;
    OC1CONbits.OCTSEL = 0;
    T2CONbits.TCKPS = 0b010;          // set prescale value to 4
    T2CONbits.ON = 1;          // Enable Timer2
    OC1CONbits.ON = 1;         // Enable OC1
}

void INT1_init(void){ // DOES NOT WORK YET
    
    TRISDbits.TRISD8 = 1;
    IEC0bits.INT1IE = 1;
    IPC1bits.INT1IP = 0b111;
    IFS0bits.INT1IF = 0;
    INTCONbits.INT1EP = 1;
    // Read port
    int val = PORTD;
}

void __ISR(16, IPL6AUTO) Timer2Hand(void) //computes rpm each 1/5 of a second
{
T4IntFlag = 0; // clear flag
oldRPM = newRPM;
newRPM = 5*CN_count*10;  // Square wave frequency = 5*count. RPM = 20* sq Wave frequency
error = desiredRPM - newRPM;
if(integral > 60000){
    integral = 60000;
}
else{
    if(newRPM>1000){
     integral = (integral + error/5); //integral in kRPM*s
    }
    else
        integral = 0;
}
derivative = (newRPM - oldRPM)*5; //derivative in RPM/s
CN_count=0;
}

void __ISR(7, IPL7SOFT) INT1_ISR(void){
    int val = PORTD;
    CN_count++;
    IFS0bits.INT1IF= 0;
}
void initt4(void){
T4CONbits.ON = 0; // turn it off
TMR4 = 0; // clear any count
T4CONbits.TCKPS = 0b101; // divide by 32
PR4 = 62195; // 32*62500 = 2,000,000 // 1/10Mhz = .0000001*PR*PreScale = .20
}


#define T4ON T4CONbits.ON

void init_PID_RPM(){
     init_OC();
     integral = 0;
    IEC0bits.T4IE = 1; // enable timer2 interrupt
    IPC4bits.T4IP = 6; //int priority -- must be greater than 0
    // initialize timer 4 //
    initt4();
    INT1_init();
    T4ON = 1; // turn on timer
}

void updateOC(){
    OC1RS=PID;
    float errorF= (float) error, integralF = (float) integral, derivativeF = (float) derivative;
    PID = servoMin - errorF*Kp - integralF*Ki + derivativeF*Kd - .5;

    if(PID > servoMin){
        PID = servoMin;
    }else if(PID < servoMax){
        PID = servoMax;
    }
}

/* 
 * 
 * 
 * 
 * 
 * CAN Functions 
 * 
 * 
 * 
 * 
 */

void dotheCAN(void){
    initCAN2();
    initFIFO();
    Read_FIFO();
    
}

void initCAN2(void){
    
    /* Turn CAN off to Reset it */
    C2CONbits.REQOP = 4;    /* Place the CAN module in Configuration mode. */
    while(C2CONbits.OPMOD != 4);
    
    C2CONCLR = 0x00008000;        /* Switch the CAN module off by clearing the ON bit */
    while(C2CONbits.CANBUSY == 1);
    
    /* Switch the CAN module back ON and switch it to Configuration mode. Wait till the switch is complete */
    C2CONbits.ON = 1;
    C2CONbits.REQOP = 4;
    while(C2CONbits.OPMOD != 4);
    
    /* Configure the CAN Module Clock */
    set_Baud();
    
    /* Set CAN2_RX pin as input */
    AD1PCFG = 0x00004100;       /* Disable Analog Mode */
    TRISBbits.TRISB14 = 1;
    
    /* Put it back in normal mode */
    C2CONbits.REQOP = 0;
    while(C2CONbits.OPMOD != 0);

    /* Configure filters and mask */
    initFilter();
    
}

void set_Baud(void){
/* Example 34-17: Configuring the CAN Module to Obtain a Specific Bit Rate */

/* This code example shows how to configure the CAN module to obtain a */
/* specific bit rate implementing the configuration shown in Example 34-16. */
/* Fsys = System Clock Frequency = 80MHz; */
/* Fbaud = CAN bit rate = 250000; */
/* N = Time Quanta (Tq) per bit = 16; */
/* Prop Segment = 1 Tq */
/* Phase Seg 1 = 7 Tq */
/* Phase Seg 2 = 7 Tq */
/* Sync Jump Width = 2 Tq */

/* Ensure the CAN module is in configuration mode.*/

    C2CONbits.REQOP = 4;
    while(C2CONbits.OPMOD != 4);

    C2CFGbits.SEG2PHTS = 1;     /* Phase seg 2 is freely programmable */
    C2CFGbits.SEG2PH = 6;       /* Phase seg 2 is 7 Tq.*/
    C2CFGbits.SEG1PH = 6;       /* Phase seg 1 is 7 Tq.*/
    C2CFGbits.PRSEG = 0;        /* Propagation seg 2 is 1 Tq. */
    C2CFGbits.SAM = 0;          /* Sample bit 1 time. */
    C2CFGbits.SJW = 2;          /* Sync jump width is 2 Tq */
    C2CFGbits.BRP = 9;         /* BRP is (Fsys/(2*Ftq))-1 = (80e6/(2*4e6))-1 = 9 */
}

void initFIFO(){
/* Allocate a total of 16 words */
    unsigned int CanFifoMessageBuffers[32];

/* Request CAN to switch to configuration mode and wait until it has switched */
    C2CONbits.REQOP = 4;
    while(C2CONbits.OPMOD != 4);

/* Initialize C2FIFOBA register with physical address of CAN message Buffer */
    C2FIFOBA = KVA_TO_PA(CanFifoMessageBuffers);

/* Configure FIFO0. This will be a receive FIFO for 4 full messages */
    C2FIFOCON0bits.TXEN = 0;    /* Clear the TXEN bit */
    C2FIFOCON0bits.DONLY = 0;   /* full message */
    C2FIFOCON0bits.FSIZE = 3;   /* size is 4 message buffers (MB0 to MB3) */
    
/* Configure FIFO1. This will be a receive FIFO for 4 full messages */
    C2FIFOCON1bits.TXEN = 0;    /* Clear the TXEN bit */
    C2FIFOCON1bits.DONLY = 0;   /* full message */
    C2FIFOCON1bits.FSIZE = 3;   /* size is 4 message buffers (MB0 to MB3) */

/* The CAN module can now be placed into normal mode if no further
 *  configuration is required. */
    C2CONbits.REQOP = 0;
    while(C2CONbits.OPMOD != 0);
}

void Read_FIFO(void){
/* From Example 34-4: Reading Received Messages from the FIFO */
/* FIFO0 size is 8 messages and each message is 4 words long. */
    unsigned int * currentMessageBuffer; /* Points to message buffer to be read */
    while(C2CONbits.ON == 1){
    /* MESSAGE 1 */
        /* Keep reading until the FIFO0 is empty. */
        while(C2FIFOINT0bits.RXNEMPTYIF == 1){
            /* Get the address of the message buffer to read from the C2FIFOUA0
             * register. Convert this physical address to virtual address. */
                currentMessageBuffer = PA_TO_KVA1(C2FIFOUA0);
                ProcessMessage1(currentMessageBuffer);
            /* Set the UINC bit to tell the CAN module that
             * a message has been read. */
                C2FIFOCON0bits.UINC = 1;
        }
        
    /* MESSAGE 2 */
    /* Keep reading until the FIFO1 is empty. */
        while(C2FIFOINT1bits.RXNEMPTYIF == 1){
            /* Get the address of the message buffer to read from the C2FIFOUA0
             * register. Convert this physical address to virtual address. */
                currentMessageBuffer = PA_TO_KVA1(C2FIFOUA1);
                ProcessMessage2(currentMessageBuffer);
            /* Set the UINC bit to tell the CAN module that
             * a message has been read. */
            C2FIFOCON1bits.UINC = 1;
        }
    }
}

void ProcessMessage1(int currentMessageBuffer){
    CANRxMessageBuffer *buffer;
    
/* When a message have been received and read, the individual fields of */
/* the received message can be queried as such. */
    buffer = (CANRxMessageBuffer *) (PA_TO_KVA1(C2FIFOUA0));
    
    if(buffer->CMSGEID.DLC == 8){
    /* If the length of the received message is 8 then process the message. */
        unsigned int * bufferToRead;
    /* Check if there is a message available to read. */
        if(C2FIFOINT0bits.RXNEMPTYIF == 1){
        /* Get the address of the buffer to read */
            bufferToRead = PA_TO_KVA1(C2FIFOUA0);
            
            EID = buffer->CMSGEID.EID;
            int D1 = buffer->CMSGDATA0.Byte0;
            int D2 = buffer->CMSGDATA0.Byte1;
            float D3 = buffer->CMSGDATA0.Byte2;
            float D4 = buffer->CMSGDATA0.Byte3;

            int D5 = buffer->CMSGDATA1.Byte0;
            int D6 = buffer->CMSGDATA1.Byte1;
            int D7 = buffer->CMSGDATA1.Byte2;
            int D8 = buffer->CMSGDATA1.Byte3;

            rpm = (D2*256)+D1;
            motor_current = ((D4*256)+D3)/10;
            bat_voltage = ((D6*256)+D5)/10;
            error_lsb = D7;
            error_msb = D8;
            
            // Store Full message 1 //
            msg_array[0] = buffer->messageWord[0];
            msg_array[1] = buffer->messageWord[1];
            msg_array[2] = buffer->messageWord[2];
            msg_array[3] = buffer->messageWord[3];

        /* Update the message buffer pointer. */
            C2FIFOCON0bits.UINC = 1;
            
        /* Turn CAN off */
            C2CONbits.REQOP = 4;    /* Place the CAN module in Configuration mode. */
            while(C2CONbits.OPMOD != 4);
            C2CONbits.ON = 0;
            while(C2CONbits.ON != 0);
        }
    }
}

void ProcessMessage2(int currentMessageBuffer){
    CANRxMessageBuffer *buffer;
    
/* When a message have been received and read, the individual fields of */
/* the received message can be queried as such. */
    buffer = (CANRxMessageBuffer *) (PA_TO_KVA1(C2FIFOUA1));
    
    if(buffer->CMSGEID.DLC == 8){
    /* If the length of the received message is 8 then process the message. */
        unsigned int * bufferToRead;
    /* Check if there is a message available to read. */
        if(C2FIFOINT1bits.RXNEMPTYIF == 1){
        /* Get the address of the buffer to read */
            bufferToRead = PA_TO_KVA1(C2FIFOUA1);
            
            EID = buffer->CMSGEID.EID;
            float D1 = buffer->CMSGDATA0.Byte0;
            int D2 = buffer->CMSGDATA0.Byte1;
            int D3 = buffer->CMSGDATA0.Byte2;
            int D4 = buffer->CMSGDATA0.Byte3;

            int D5 = buffer->CMSGDATA1.Byte0;
            int D6 = buffer->CMSGDATA1.Byte1;
            int D7 = buffer->CMSGDATA1.Byte2;
            int D8 = buffer->CMSGDATA1.Byte3;
            
            throttle = (D1/255)*100;
            contr_temp = D2 - 40;
            motor_temp = D3 - 30;
            contr_status = D5;
            switch_status = D6;
            
            // Store Full message 2 //
            msg_array[4] = buffer->messageWord[0];
            msg_array[5] = buffer->messageWord[1];
            msg_array[6] = buffer->messageWord[2];
            msg_array[7] = buffer->messageWord[3];

        /* Update the message buffer pointer. */
            C2FIFOCON1bits.UINC = 1;
            
        /* Turn CAN off */
            C2CONbits.REQOP = 4;    /* Place the CAN module in Configuration mode. */
            while(C2CONbits.OPMOD != 4);
            C2CONbits.ON = 0;
            while(C2CONbits.ON != 0);
        }
    }
}

void initFilter(void){
    /* Filter 0 is set up to accept messages with Kelly controller EID. 
     * Accepted messages will be stored in FIFO0. Mask 0 is
     * used to implement the filter address range. */
    
    /* Check that it's in in configuration mode */
    C2CONbits.REQOP = 4;
    while(C2CONbits.OPMOD != 4);    
    
    /* Kelly Message 1 Filter */
    
    C2FLTCON0bits.FSEL0 = 0x0000;           /* Store messages in FIFO0 */
    C2FLTCON0bits.MSEL0 = 0x00;             /* Use Mask 0 */
    
    C2RXF0bits.SID = 0b10011110000;         /* Kelly SID 0x0CF0*/
    C2RXF0bits.EID = 0b010001111000000101;  /* Kelly message 1 EID 0x11E05*/
    
    C2RXF0bits.EXID = 1;                    /* Match only messages with extended identifier addresses */
    C2RXM0bits.SID = 1;                     /* Allow only Kelly's SID bits */
    C2RXM0bits.EID = 0b111111111111111111;  /* Allow only 0x11E05 EID bits */
    C2RXM0bits.MIDE = 1;                    /* Match only EID message types. */
    
    C2FLTCON0bits.FLTEN0 = 1;               /* Enable the filter */
    
    /* Message 2 Filter */
    
    C2FLTCON0bits.FSEL1 = 0x0001;           /* Store messages in FIFO1 */
    C2FLTCON0bits.MSEL1 = 0x00;             /* Use Mask 0 (already configured) */
    
    C2RXF1bits.SID = 0b10011110000;         /* Kelly SID 0x0CF0*/
    C2RXF1bits.EID = 0b010001111100000101;  /* Kelly message 2 EID 0x11F05*/
    
    C2RXF1bits.EXID = 1;                    /* Match only messages with extended identifier addresses */
    
    C2FLTCON0bits.FLTEN1 = 1;               /* Enable the filter */
}

int last_Kelly_msg = 0;
void whichKelly(void){
    if (msg_array[5] == 0x41){
        // message from G Kelly
        mot_G = motor_current;
        contr_temp_G = contr_temp;
        mot_temp_G = motor_temp;
    }
    else if(msg_array[6] == 0x1E){
        // message from one of the motor Kellys
        mot_last = mot_new;
        mot_new = motor_current;
        
        rpm_last = rpm_new;
        rpm_new = rpm;
        
        contr_temp_LR = contr_temp;
        mot_temp_LR = motor_temp;
    }
}

void computeDynamics(void){
    whichKelly();
    avg_rpm = (rpm_last + rpm_new)/2;
    speed = avg_rpm/16.4;       // speed in mph with wheel radius of 10.25 inches
    
    if(BRAKE > BRAKE_OUT_MIN){
        power = mot_last + mot_new - mot_G;
    }
    else{
        power = -mot_last-mot_new-mot_G;
    }
}

void decodeErrors(void) {
    
    uint16_t error_msg = (error_msb << 8) | (error_lsb & 0xff);
    int i=0;
    int comparison = 0b00000001;
    char *Error_messages[] = {"ID Angle", 
                                "Over Voltage", 
                                "Low Voltage",
                                "", 
                                "Stall", 
                                "Internal Volts Fault",
                                "Controller temp",
                                "H Pedal",
                                "",
                                "Reset",
                                "Throttle Fault",
                                "Angle sensor",
                                "",
                                "",
                                "Motor Temp",
                                "Hall Galvanometer"};


    char* messages = " "; 
    new_CAN_Error_MSG = 0;
    CAN_COUNT=0;
    for(i=0;i<16; i++){
        if((error_msg & comparison) > 0){
            errorPointer[CAN_COUNT] = Error_messages[i];
            CAN_COUNT++;
            new_CAN_Error_MSG = 1;
        }
        comparison = comparison << 1;
    }
    
    
}

/* 
 * 
 * 
 * 
 * 
 * SPI Functions 
 * 
 * 
 * 
 * 
 */

void initDAC_SPI2(void){
    
    T5CONbits.ON = 0;           // Turn off timer2 so as not to interrupt Rotary Encoder
    
    SPI2CONbits.ON = 0;         // Turn off SPI module
    
    CS_G = 1;                   // Set CS_G low (idle state)
    CS_L = 1;                   // Set CS_L low (idle state)
    CS_R = 1;                   // Set CS_R low (idle state)
    
    DDPCON = 0x0000;            // disable JTAG
    IEC1bits.SPI2EIE = 0;       // SPI interrupts disabled
    IEC1bits.SPI2RXIE = 0;
    IEC1bits.SPI2TXIE = 0;
    
    SPI2BUF = 0;                // Clear the receive buffer
     
    SPI2BRG = 5;                // FSCK = 10MHz/(2*(BRG+1)) = 0.83MHz
     
    SPI2STATbits.SPIROV = 0;    // Clear overflow flag
    
    // SPI2CON settings for DAC //
    SPI2CONbits.FRMEN = 0;      // Framed SPI support is disabled
    SPI2CONbits.SIDL = 0;       // Continue operation in IDLE mode
    SPI2CONbits.DISSDO = 0;     // SDO1 pin is controlled by the module
    SPI2CONbits.ENHBUF = 0;     // Disable enhanced buffer mode
    SPI2CONbits.MODE16 = 1;     // 16 bit mode
    SPI2CONbits.CKP = 1;        // Idle state for clock is high, active state is low, on rising edge
    SPI2CONbits.CKE = 0;        // Output data changes on transition from idle to active
    SPI2CONbits.SSEN = 0;       // Not in slave mode
    SPI2CONbits.MSTEN = 1;      // Master mode
    SPI2CONbits.SMP = 0;        // Input data sampled at the middle of data output time
    
    SPI2CONbits.ON = 1;         // Turn module on
    
}

void initENCODER_SPI2(void){
    SPI2CONbits.ON = 0;         // Turn off SPI module
    CS_ROT = 1;                   // Set CS_ROT high (idle state)
    
    DDPCON = 0x0000;            // disable JTAG
    IEC1bits.SPI2EIE = 0;       // SPI interrupts disabled
    IEC1bits.SPI2RXIE = 0;
    IEC1bits.SPI2TXIE = 0;
    
    CNPUEbits.CNPUE9 = 1;       // pull up 
    CNPUEbits.CNPUE10 = 1;
    
    SPI2BUF = 0;                // Clear the receive buffer
     
    SPI2BRG = 5;                // FSCK = 10MHz/(2*(BRG+1)) = 0.83MHz
     
    SPI2STATbits.SPIROV = 0;    // Clear overflow flag
    
    // SPI2CON settings for Encoder //
    SPI2CONbits.FRMEN = 0;      // Framed SPI support is disabled
    SPI2CONbits.SIDL = 0;       // Continue operation in IDLE mode
    SPI2CONbits.DISSDO = 0;     // SDO1 pin is controlled by the module
    SPI2CONbits.ENHBUF = 0;     // Disable enhanced buffer mode
    SPI2CONbits.MODE16 = 0;     // 16 bit mode OFF
    SPI2CONbits.CKP = 0;        // Idle state for clock is low, active state is high, on rising edge
    SPI2CONbits.CKE = 1;        // Output data changes on transition from active to idle
    SPI2CONbits.SSEN = 0;       // Not in slave mode
    SPI2CONbits.MSTEN = 1;      // Master mode
    SPI2CONbits.SMP = 0;        // Input data sampled at the middle of data output time
    
    SPI2CONbits.ON = 1;         // Turn module on
    
}

void initISO_SPI2(void){
    CS_ISOB = 1;                // Set CS high (idle state)
    DDPCON = 0x0000;            // disable JTAG
    IEC1bits.SPI2EIE = 0;       // SPI interrupts disabled
    IEC1bits.SPI2RXIE = 0;
    IEC1bits.SPI2TXIE = 0;
    SPI2CONbits.ON = 0;         // Turn off SPI module
    
    CNPUEbits.CNPUE9 = 1;
    CNPUEbits.CNPUE10 = 1;
     
    SPI2BUF = 0;                // Clear the receive buffer
    SPI2BRG = 19;               // FSCK = 1MHz
    
    SPI2STATbits.SPIROV = 0;    // Clear overflow flag
    
     
    // SPI2CON settings for AMS //
    SPI2CONbits.FRMEN = 0;      // Framed SPI support is disabled
    SPI2CONbits.SIDL = 0;       // Continue operation in IDLE mode
    SPI2CONbits.DISSDO = 0;     // SDO1 pin is controlled by the module
    SPI2CONbits.CKP = 1;        // Idle state for clock is high, active state is low, on rising edge
    SPI2CONbits.CKE = 0;        // Output data changes on transition from idle to active
    SPI2CONbits.SSEN = 0;       // Not in slave mode
    SPI2CONbits.MSTEN = 1;      // Master mode
    SPI2CONbits.SMP = 0;        // Input data sampled at the middle of data output time
     
    SPI2CONbits.ON = 1;         // Turn module on
}

short WriteReadSPI(unsigned short i){
    SPI2BUF = i;                    // Write to buffer for transmission
    while (!SPI2STATbits.SPIRBF);   // Wait for transfer to be completed
    return SPI2BUF;                 // Return the received value
}

short ReadSPI(unsigned short i){
    SPI2BUF = i;                    // Write to buffer for transmission
    while (!SPI2STATbits.SPIRBF);   // Wait for transfer to be completed
    unsigned short x = SPI2BUF;
    return SPI2BUF;                 // Return the received value
}

/*
 * 
 * 
 * 
 * Interrupt Timer Configurations
 * 
 * 
 * 
 * 
 */

/* Configure timer2 to kill 20us (For Rotary Encoder) */
void initTimer5(void){              // initialize 100ms timer on timer 2
    
    T5CONbits.ON = 0;               // set timer off
    
    /* Set timer length for 100 ms
       * PR*PreScale = 256*31250 = 8M
       * 80Mhz/(PR*PreScale) = 10 Hz*/
    T5CONbits.TCKPS = 0b111;
    PR5 = 31250;
    
    TMR5 = 0;                       // reset timer 2
    
    IPC5bits.T5IP = 0b101;          // Set priority level = 5
    IFS0bits.T5IF = 0;              // Clear interrupt flag
    IEC0bits.T5IE = 1;              // Enable interrupt
}

/* Timer 5 Interrupt Service Routine */
/*vector number for T5 is 20, Priority 5 */
void __ISR(20, IPL5SOFT) _Timer5ISR(void){
    
    IFS0bits.T5IF = 0;                   // clear timer flag
    if(SPI_init_done==1){
        initENCODER_SPI2();
        SPI_init_done=0;
    }
   
    if(Encodertransaction==1){
        switch (read_state){
            case 0:
                /* Set timer length
                * PR*PreScale = 64*25 = 
                * 80Mhz/(PR*PreScale) = 50,000 Hz*/
                T5CONbits.ON=0;
                T5CONbits.TCKPS = 0b110;
                PR5 = 25;
                TMR5 = 0;
                T5CONbits.ON = 1;
                read_state = 1;
                break;
            case 1:                         // send read position command to encoder
                CS_ROT = 0;
                readbyte = WriteReadSPI(0x10);
                read_state = 2;
                CS_ROT = 1;
                break;
            case 2:                         // check if encoder responded with read position
                if (readbyte != 0x10){  
                    CS_ROT = 0;
                    readbyte = WriteReadSPI(0x00);
                    CS_ROT = 1;
                    break;
                }
                if (readbyte == 0x10){
                    read_state = 3;
                }
                break;
            case 3:                         // read in MSB of encoder position
                CS_ROT = 0;
                readMSB = WriteReadSPI(0x00);
                read_state = 4;
                CS_ROT = 1;
                break;
            case 4:                         // read in LSB of encoder position
                CS_ROT = 0;
                readLSB = WriteReadSPI(0x00);
                T5CONbits.ON = 0;
                
                /* Set timer length
                * PR*PreScale = 64*25 = 
                * 80Mhz/(PR*PreScale) = 50,000 Hz*/
                T5CONbits.TCKPS = 0b110;
                PR5 = 25;
                TMR5 = 0;
                T5CONbits.ON = 1;
                read_state = 0;
                first_time = 0;
                CS_ROT = 1;
                my_MSB = (readMSB << 8);
                my_LSB = readLSB;
                my_Position = (my_MSB | my_LSB);

                steering_angle = degrees*my_Position;
                steering_angle = steering_angle/positions;
                Encodertransaction=0;
                break;
        }
    }
}

void initTimer3(void){
   
    // Initialize timer 3 interrupt bits //
    IEC0bits.T3IE = 1; // enable timer3 interrupt
    IPC3bits.T3IP = 0b100; //int priority -- 4

    // Enable timer 3 //
    T3CONbits.ON = 0; // turn it off
    TMR3 = 0; // clear any count
    
    /* Set timer length
     * PR*PreScale = 64*2500 = 320,000
     * 80Mhz/(PR*PreScale) = 250 Hz */
    PR3 = 5000;
    T3CONbits.TCKPS = 0b110; // divide by 64
    
    T3CONbits.ON = 1; // turn on timer3
    
}

/* Timer 3 Interrupt Service Routine */
/* vector number for T3 is 12, Priority 6 */
void __ISR(12, IPL6AUTO) Timer3Hand(void){
    //count up to certain time for charging caps before entering next drive state
    if(drive_state == Precharge){
        T_charge = T_charge+1;
    }
    
    if (AIR_INTERRUPT){ // Check if Kellys are powered
        dotheCAN();
        computeDynamics();
        decodeErrors();
    }
    
    IFS0bits.T3IF = 0; // clear interrupt flag
}

void initInterrupts(void){
    INTCONbits.MVEC = 1;        // enable multivectored interrupts
    asm("ei");                  // enable interrupts
    initTimer5();               // initialize timer for Rotary Encoder
    initTimer3();               // Initialize CAN timer
    T5CONbits.ON = 1;           // turn on timer
}

/*
 * 
 * 
 * 
 * IO Functions
 * 
 * 
 * 
 * 
 */

void initIO(void){
    
    TRISEbits.TRISE7 = 0;   // Rotary Encoder CS set as output
    TRISEbits.TRISE6 = 0;   // GEN DAC CS set as output
    TRISGbits.TRISG9 = 0;   // Left DAC CS set as output
    TRISFbits.TRISF3 = 0;   // Right DAC CS set as output
    
    TRISBbits.TRISB2 = 1;       // configure analog pin RB2 (direction_in) as input
    TRISEbits.TRISE3 = 0;       // configure digital pin RE3 (direction_out) as output
    TRISCbits.TRISC13 = 0;      // configure digital pin RC13 (AIR_interrupt) as output
    TRISBbits.TRISB4 = 1;       // configure pin RB4 (throttle1) as input
    TRISBbits.TRISB5 = 1;       // configure pin RB5 (throttle2) as input
    TRISBbits.TRISB3 = 1;       // configure pin RB3 (brake) as input
    
    TRISDbits.TRISD7 = 1;       // configure pin RD7 (Enable) as input
    TRISDbits.TRISD6 = 1;       // configure pin RD6 (Charge Enable) as input
    
    TRISEbits.TRISE5 = 0;       // configure digital pin RE5 (ISOB) as output

    
    //precharge and discharge pins
    TRISEbits.TRISE2 = 0;
    TRISDbits.TRISD11 = 0;
    
    // Set all inputs to digital //
    AD1PCFG = 0xFFFF;
    
    // Enable analog control for CAN2 pins //
    AD1PCFG = 0x00004100;
    
    // Enable analog pins for pedal inputs //
    AD1PCFGbits.PCFG2 = 0;  // Turn on AN2 for pin RB2 (direction_in)
    AD1PCFGbits.PCFG4 = 0;  // Turn on AN4 for pin RB4 (Throttle1)
    AD1PCFGbits.PCFG5 = 0;  // Turn on AN5 for pin RB5 (Throttle2)
    AD1PCFGbits.PCFG3 = 0;  // Turn on AN3 for pin RB3 (Brake)
    
    // Turn all contactors OFF //
    PRECHARGE_CONTACTOR = 0;
    DISCHARGE_CONTACTOR = 0;
    AIR_INTERRUPT = 0;
    
    // Enable pull up resistors //
    CNPUEbits.CNPUE15 = 1;  // configure pull up resistor for ENABLE pin
    CNPUEbits.CNPUE16 = 1;  // configure pull up resistor for CHARGE_ENABLE pin
    
    //configure ignition 
    TRISBbits.TRISB6 = 0;
    
}

void write_to_LDAC(int Throttle, int Brake){
    initDAC_SPI2();
    CS_L=0;
    WriteReadSPI(Throttle); // throttle
    CS_L=1;
    Nop();
    Nop();
    CS_L=0;
    WriteReadSPI(Brake); // brake
    CS_L=1;
    Nop();
    Nop();
 
}

void write_to_RDAC(int Throttle, int Brake){
    CS_R=0;
    WriteReadSPI(Throttle); // throttle
    CS_R=1;
    Nop();
    Nop();
    CS_R=0;
    WriteReadSPI(Brake); // brake
    CS_R=1;
    Nop();
    Nop();
}

void write_to_GDAC(int Throttle, int Brake){
    CS_G=0;
    WriteReadSPI(Throttle); // throttle
    CS_G=1;
    Nop();
    Nop();
    CS_G=0;
    WriteReadSPI(Brake); // brake
    CS_G=1;
    Nop();
    Nop();
}

void initadc(void){
    AD1CON1bits.FORM = 0; // 16 bit unsigned integer output
    AD1CON1bits.ASAM = 0; // turn off auto sampling
    AD1CON2bits.VCFG = 0; // Avdd and Avss references
    AD1CON2bits.CSCNA = 0; // turn off scan mosd
    AD1CON2bits.BUFM = 0; // single buffer
    AD1CON2bits.ALTS = 0; // just use mux a
    AD1CON3bits.ADRC = 0; // use PB clock not RC
    AD1CON3bits.ADCS = 7; // divide pbclk by(works up to 80 MHz)
    AD1CHSbits.CH0NA = 0; // standard reverence (not an1)
    AD1CON1bits.ON = 1;   // turn it on
}

unsigned short convert_to_analog(unsigned short AN_PIN){
    AD1CHSbits.CH0SA = AN_PIN;  // select channel
    SAMP = 1;                   // sample
    // delay //
    int i = 0;
    while (i < 10){
        i++;
    }
    SAMP = 0;                   // start conversion
    while (!ADCDONE);           // wait for completion
    return ADC1BUF0;            // return result
}

/*
 * 
 * 
 * 
 * UART Functions
 * 
 * 
 * 
 * 
 */

void configUART2(void){
    U2MODEbits.ON = 0;          // be sure it is off
    U2MODEbits.BRGH = 1;        // high speed mode
    U2BRG = 42;                // store value in register
    
    U2MODEbits.PDSEL = 0;       // 8 bit data, no parity
    U2MODEbits.STSEL = 0;       // 1 stop bit
    
    U2MODEbits.UEN = 0b00;      // UxTX and UxRX pins are enabled and used; no flow control pins
    U2STAbits.UTXEN = 1;        // enable transmitter
    U2STAbits.URXEN = 1;        // enable receiver
    U2MODEbits.ON = 1;          // turn it on
    
}

#define data_avail U2STAbits.URXDA // data avail
unsigned char tgetc(void){
    unsigned char val;
    while (!data_avail); // wait for character to appear
    val = U2RXREG; // get character
    tputc(val); // echo
    return val;
}

#define buffer_full U2STAbits.UTXBF // buffer full
void tputc(unsigned char val){
    while (buffer_full); // wait if buffer full
    U2TXREG = val; // send character
    return;
}

void _mon_putc(char c){
    tputc(c);
}

/*
 * 
 * 
 * 
 * Begin Main
 * 
 * 
 * 
 * 
 */

int main(int argc, char** argv) {
    
    // Configure IO //
    initIO();
  
    // Configure UART //
    configUART2();
    
    /* Turn on ADC */
    initadc();
    
    // Turn all DACs ON //
    write_to_LDAC(0x1000, 0x1000);
    write_to_GDAC(0x1000, 0x1000);
    write_to_RDAC(0x1000, 0x1000);
    
    // Initialize Interrupt Timers //
    initInterrupts();
    init_PID_RPM();
    OC1CONbits.ON = 0;
    
    // Configure LCD //
    initLCD();
    
    // Start Drive States Loop //
    while(1){
           
        /* Convert analog pin values */
        AN_BRAKE = convert_to_analog(BRAKE);
        AN_THROTTLE1 = convert_to_analog(THROTTLE1);
        AN_THROTTLE2 = convert_to_analog(THROTTLE2);
        AN_DIRECTION_IN = convert_to_analog(DIRECTION_IN);
    
        throttle1_val = ((float)AN_THROTTLE1-THROTTLE1_OUT_MIN)*THROTTLE1_CONST;
        throttle2_val = ((float)AN_THROTTLE2-THROTTLE2_OUT_MIN)*THROTTLE2_CONST;
        
        // THIS SECTION IS USED TO CALCULATE THE THROTTLE AND BRAKE INPUT VALUES //
        
            //shutdown systems if more than 5% difference in throttle input values, or throttle values outside ranges
        if (abs(throttle1_val-throttle2_val)>=THROTTLE_DIFF || AN_THROTTLE1<THROTTLE1_OUT_MIN || AN_THROTTLE1>THROTTLE1_OUT_MAX
                || AN_THROTTLE2<THROTTLE2_OUT_MIN || AN_THROTTLE2>THROTTLE2_OUT_MAX){ //Issue was resolved. It was THROTTLE MIN
            AIR_INTERRUPT = 0;      //Turn off AIR Relay
            throttle_val_final = 0;
            updateScreen();
            Error=1;
        }
            //calculate throttle value from pedal 
        else {
            throttle_val_final = (throttle1_val+throttle2_val)/2;
        }
        
            //if value outside range, trigger error and shutdown    
        if (AN_BRAKE<BRAKE_OUT_MIN || AN_BRAKE>BRAKE_OUT_MAX){
            AIR_INTERRUPT = 0;  //Turn off AIR relay
            brake_val_final = 0;
            updateScreen();
            Error=1;
        }
           //calculate brake value to send to the Kelly controller  
        else{
            brake_val_final = (AN_BRAKE-BRAKE_OUT_MIN)*BRAKE_CONST;      
        }
        
        //establish state diagram(s)
        //charge states switch
        switch(charge_state) {
            case Idle:
                charge_state_LCD = "Idle";
               OC1CONbits.ON = 0;
               initDAC_SPI2();
               write_to_GDAC(0xB000+KELLY_DEAD_MIN, 0x3000+KELLY_DEAD_MIN);  
                    Encodertransaction = 1;
                    SPI_init_done=1;
                    T5CONbits.ON=1;
                if (Error == 1) {
                    drive_state = Wait;
                }
                else if((!CHARGE_ENABLE) && (endSum < MinV) && (newRPM > 1000) ) {
                    charge_state = Charge;
                }
                break;
            case Charge:
                charge_state_LCD = "Charging";
                if(!OC1CONbits.ON){
                    init_PID_RPM();
                }
                updateOC();
                initDAC_SPI2();
                write_to_GDAC(0xB000+KELLY_DEAD_MIN, 0x3000+KELLY_DEAD_MIN+REGEN_LEVEL);  
                Encodertransaction = 1;
                SPI_init_done=1;
                T5CONbits.ON=1;
                if (Error == 1) {
                    drive_state = Wait;
                }
                else if ((endSum > MaxV) || (newRPM < 1000) || (CHARGE_ENABLE)) {
                    charge_state = Idle;
                }
                break;
        }
        if (ENABLE) {
            drive_state = Wait;
            T_charge = 0;
            T3CONbits.ON = 1;
        }
      
        switch(drive_state){
            case Wait:
                direction_state_LCD = "Wait";
                IGNITION = 0;
                PRECHARGE_CONTACTOR = 0;
                DISCHARGE_CONTACTOR = 0;
                AIR_INTERRUPT = 0;
                charge_state = Idle;
                //set signals to send to Kellys
                Throttle_L = KELLY_DEAD_MIN;
                Throttle_R = KELLY_DEAD_MIN;
                Brake_L = KELLY_DEAD_MIN;
                Brake_R = KELLY_DEAD_MIN;
            
                // Read AMS Info //
                getInfo();
                endSum = ((uint16_t) sumAll)*100*30/1000000 - .5;
                
                if(Encodertransaction == 0){
                    // Send Neutral Throttle and Brake to LDAC and RDAC //
                    write_to_LDAC(0xB000+Throttle_L, 0x3000+Brake_L);
                    write_to_RDAC(0xB000+Throttle_R, 0x3000+Brake_R);
                    //as soon as you turn on Kellys, generator gets dead minimum value
                    write_to_GDAC(0xB000+KELLY_DEAD_MIN, 0x3000+KELLY_DEAD_MIN);  
                    Encodertransaction = 1;
                    SPI_init_done=1;
                    T5CONbits.ON=1;
                    
                }
                
                updateScreen();
                
                if (Error == 1) {
                    drive_state = Wait;
                }
                else if ((AN_DIRECTION_IN >= FWD_MIN && AN_DIRECTION_IN <= FWD_MAX) || 
                    (AN_DIRECTION_IN >= REV_MIN && AN_DIRECTION_IN <= REV_MAX) || (ENABLE == 1)){
                    drive_state = Wait;
                }
                else{
                    drive_state = Precharge;
                }
                updateScreen();
                break;
            case Precharge:
                direction_state_LCD = "Precharge";
                PRECHARGE_CONTACTOR = 1;
                DISCHARGE_CONTACTOR = 1;
                AIR_INTERRUPT = 0;
            
                //MUST INTEGRATE ERROR CODE
                if (Error == 1) {
                    drive_state = Wait;
                }
                //PLACE HOLDER TIME COUNT
                else if (T_charge >= 140) {
                    drive_state = Neutral;
                }
                else {
                    charge_state = Idle;   
                }
                updateScreen();
                break;
            case Neutral:
                direction_state_LCD = " Neutral ";
                IGNITION = 1;
                DISCHARGE_CONTACTOR = 1;
                AIR_INTERRUPT = 1;  
                T_charge=0;
                DIRECTION_OUT = 0;
                
                //set signals to send to Kellys
                Throttle_L = KELLY_DEAD_MIN;
                Throttle_R = KELLY_DEAD_MIN;
                Brake_L = KELLY_DEAD_MIN;
                Brake_R = KELLY_DEAD_MIN;
            
                // Read AMS Info //
                getInfo();
                endSum = ((uint16_t) sumAll)*100*30/1000000 - .5;
                
                if(Encodertransaction == 0){
                    // Send Neutral Throttle and Brake to LDAC and RDAC //
                    write_to_LDAC(0xB000+Throttle_L, 0x3000+Brake_L);
                    write_to_RDAC(0xB000+Throttle_R, 0x3000+Brake_R);
                    Encodertransaction = 1;
                    SPI_init_done=1;
                    T5CONbits.ON=1;
                }
                if (Error == 1) {
                    drive_state = Wait;
                    updateScreen();
                }
                else if (AN_DIRECTION_IN >= FWD_MIN && AN_DIRECTION_IN <= FWD_MAX) {
                    drive_state = Fwd;
                    
                }
                else if (AN_DIRECTION_IN >= REV_MIN && AN_DIRECTION_IN <= REV_MAX) {
                    drive_state = Rev;
                    
                }
                updateScreen();
                break;
            case Fwd:
                direction_state_LCD = " Forward ";
                DIRECTION_OUT = 0;            //turn on direction relay 
                
                // Print "F" to LCD - integrate Fer's code          
                if (Error == 1) {
                    drive_state = Wait;
                    updateScreen();
                }
                
                //Left Turn
                else if (!((AN_DIRECTION_IN >= FWD_MIN && AN_DIRECTION_IN <= FWD_MAX) || 
                    (AN_DIRECTION_IN >= REV_MIN && AN_DIRECTION_IN <= REV_MAX))){
                    drive_state = Neutral;
                }
                else if (steering_angle <= 180){
                    drive_state = Left;
                }
                //Right Turn
                else if (steering_angle >= 180){
                    drive_state = Right;                    
                } 
                updateScreen();
                break;
            case Left:
                direction_state_LCD = " Forward ";

                // implement torque vectoring //
                Throttle_L=KELLY_DEAD_MIN+throttle_val_final*(1-steering_angle/360*TORQUE_VECTORING_CONSTANT);
                Throttle_R=KELLY_DEAD_MIN+throttle_val_final*(1+steering_angle/360*TORQUE_VECTORING_CONSTANT);
                Brake_L=KELLY_DEAD_MIN+brake_val_final*(1+steering_angle/360*TORQUE_VECTORING_CONSTANT);
                Brake_R=KELLY_DEAD_MIN+brake_val_final*(1-steering_angle/360*TORQUE_VECTORING_CONSTANT);
                
                //if below the minimum value
                if (Throttle_L < KELLY_DEAD_MIN) {
                    Throttle_L = KELLY_DEAD_MIN;
                }
                if (Throttle_R < KELLY_DEAD_MIN) {
                    Throttle_R = KELLY_DEAD_MIN;
                }
                if (Brake_L < KELLY_DEAD_MIN) {
                    Brake_L = KELLY_DEAD_MIN;
                }
                if (Brake_R < KELLY_DEAD_MIN) {
                    Brake_R = KELLY_DEAD_MIN;
                }
                //if above the maximum kelly value
                if (Throttle_L > KELLY_DEAD_MAX) {
                    Throttle_L = KELLY_DEAD_MAX;
                }
                if (Throttle_R > KELLY_DEAD_MAX) {
                    Throttle_R = KELLY_DEAD_MAX;
                }
                if (Brake_L > KELLY_DEAD_MAX) {
                    Brake_L = KELLY_DEAD_MAX;
                }
                if (Brake_R > KELLY_DEAD_MAX) {
                    Brake_R = KELLY_DEAD_MAX;
                }
                
                // Read AMS Info //
                getInfo();
                endSum = ((uint16_t) sumAll)*100*30/1000000 - .5;
                
                /* Send Forward State Throttle and Brake to LDAC and RDAC */
                if(Encodertransaction == 0){
                    write_to_LDAC(0xB000+Throttle_L, 0x3000+Brake_L);
                    write_to_RDAC(0xB000+Throttle_R, 0x3000+Brake_R);
                    Encodertransaction = 1;
                    SPI_init_done=1;
                    T5CONbits.ON=1;
                }
                if (Error == 1) {
                    drive_state = Wait;
                    updateScreen();
                }
                
                else if (!((AN_DIRECTION_IN >= FWD_MIN && AN_DIRECTION_IN <= FWD_MAX) || 
                    (AN_DIRECTION_IN >= REV_MIN && AN_DIRECTION_IN <= REV_MAX))){
                    drive_state = Neutral;
                }
                
                else if (steering_angle > 315){
                        drive_state = Right;
                    }
                updateScreen();
                break;
            case Right:
                direction_state_LCD = " Forward ";

                // implement torque vectoring //
                Throttle_L=KELLY_DEAD_MIN+throttle_val_final*(1+(360-steering_angle)/360*TORQUE_VECTORING_CONSTANT);
                Throttle_R=KELLY_DEAD_MIN+throttle_val_final*(1-(360-steering_angle)/360*TORQUE_VECTORING_CONSTANT);
                Brake_L=KELLY_DEAD_MIN+brake_val_final*(1-(360-steering_angle)/360*TORQUE_VECTORING_CONSTANT);
                Brake_R=KELLY_DEAD_MIN+brake_val_final*(1+(360-steering_angle)/360*TORQUE_VECTORING_CONSTANT);
                
                //if below minimum kelly value
                if (Throttle_L < KELLY_DEAD_MIN) {
                    Throttle_L = KELLY_DEAD_MIN;
                }
                if (Throttle_R < KELLY_DEAD_MIN) {
                    Throttle_R = KELLY_DEAD_MIN;
                }
                if (Brake_L < KELLY_DEAD_MIN) {
                    Brake_L = KELLY_DEAD_MIN;
                }
                if (Brake_R < KELLY_DEAD_MIN) {
                    Brake_R = KELLY_DEAD_MIN;
                }
                //if above the maximum kelly value
                if (Throttle_L > KELLY_DEAD_MAX) {
                    Throttle_L = KELLY_DEAD_MAX;
                }
                if (Throttle_R > KELLY_DEAD_MAX) {
                    Throttle_R = KELLY_DEAD_MAX;
                }
                if (Brake_L > KELLY_DEAD_MAX) {
                    Brake_L = KELLY_DEAD_MAX;
                }
                if (Brake_R > KELLY_DEAD_MAX) {
                    Brake_R = KELLY_DEAD_MAX;
                }
                
                // Read AMS Info //
                getInfo();
                endSum = ((uint16_t) sumAll)*100*30/1000000 - .5;
                
                if(Encodertransaction == 0){
                    write_to_LDAC(0xB000+Throttle_L, 0x3000+Brake_L);
                    write_to_RDAC(0xB000+Throttle_R, 0x3000+Brake_R);
                    Encodertransaction = 1;
                    SPI_init_done=1;
                    T5CONbits.ON=1;
                }
                if (Error == 1) {
                    drive_state = Wait;
                    updateScreen();
                }
                else if (!((AN_DIRECTION_IN >= FWD_MIN && AN_DIRECTION_IN <= FWD_MAX) || 
                    (AN_DIRECTION_IN >= REV_MIN && AN_DIRECTION_IN <= REV_MAX))){
                    drive_state = Neutral;
                }
                else if (steering_angle < 45)  {    
                        drive_state = Left;
                }
                updateScreen();
                break;
            case Rev:
                direction_state_LCD = " Reverse ";
                DIRECTION_OUT = 1;        // set direction relay to reverse state
                
                Rev_L=KELLY_DEAD_MIN+throttle_val_final;
                Rev_R=KELLY_DEAD_MIN+throttle_val_final;
                Brake_L=KELLY_DEAD_MIN;
                Brake_R=KELLY_DEAD_MIN;
                if (Error == 1) {
                    drive_state = Wait;
                    updateScreen();
                }
                else if (!((AN_DIRECTION_IN >= FWD_MIN && AN_DIRECTION_IN <= FWD_MAX) || 
                    (AN_DIRECTION_IN >= REV_MIN && AN_DIRECTION_IN <= REV_MAX))){
                   drive_state = Neutral;
                }
                /* Send Reverse State Throttle and Brake to LDAC and RDAC 
             (brake is at 0) */
                
                // Read AMS Info //
                getInfo();
                endSum = ((uint16_t) sumAll)*100*30/1000000 - .5;
                
                if(Encodertransaction == 0){
                    write_to_LDAC(0xB000+Rev_L, 0x3000+Brake_L);
                    write_to_RDAC(0xB000+Rev_R, 0x3000+Brake_R);
                    Encodertransaction = 1;
                    SPI_init_done=1;
                    T5CONbits.ON=1;
                }
                updateScreen();
                break;                
        }
        
        // Print to RF //
        printRF();
    }
    
    return (EXIT_SUCCESS);
}

void printRF(){
    printf("\r\n%2.2f,%d,%d,%2.2f,%3.3,%d,%d,%d,%d,%d,%d,%d",speed,newRPM,percBat,power,
            steering_angle,mot_temp_G,contr_temp_G,mot_temp_LR,contr_temp_LR,error_lsb,error_msb,switch_status);
}

